//MIESZADEŁKO MAGNETYCZNE
//Rafał Kozik
#include <avr/io.h>
#include <inttypes.h>
#include <avr/pgmspace.h>
#include <avr/delay.h>
#include <string.h>
#include <stdio.h>
#include <avr/interrupt.h>

//Obsluga LCD pochodzi z kursu programowania z Elektroniki dla Wszystkich:
// LCD_simple.c - plik pokazuj¹cy prost¹ obs³ugê wywietlacza alfanumerycznego LCD
// 
// Autor: Radosław Koppel          Kompilator: WinAVR 20050214

// Makra upraszczaj¹ce dostêp do portów
// *** Port
#define PORT(x) XPORT(x)
#define XPORT(x) (PORT##x)
// *** Pin
#define PIN(x) XPIN(x)
#define XPIN(x) (PIN##x)
// *** DDR
#define DDR(x) XDDR(x)
#define XDDR(x) (DDR##x)

// Definicje wyprowadzeñ
#define LCD_RS 0
#define LCD_RSPORT D
#define LCD_E 1
#define LCD_EPORT D
// Port danych: zak³adam, ¿e wyprowadzenia s¹ po³¹czone kolejno. 
// Definiowane jest jedynie przesuniêcie oznaczaj¹ce na którym wyprowadzeniu jest D4
#define LCD_DPORT B
#define LCD_D4 2 // D1, D2, D3 - po³¹czone kolejno


// Komendy steruj¹ce wywietlaczem
#define LCDC_CLS		0x01
#define LCDC_HOME		0x02
#define LCDC_MODE		0x04
	#define LCDC_MODER		0x02
	#define LCDC_MODEL		0
	#define LCDC_MODEMOVE	0x01
#define LCDC_ON 		0x08
	#define LCDC_ONDISPLAY	0x04
	#define LCDC_ONCURSOR	0x02
	#define LCDC_ONBLINK	0x01
#define LCDC_SHIFT		0x10
	#define LCDC_SHIFTDISP	0x08
	#define LCDC_SHIFTR		0x04
	#define LCDC_SHIFTL		0
#define LCDC_FUNC		0x20
	#define LCDC_FUNC8b		0x10
	#define LCDC_FUNC4b		0
	#define LCDC_FUNC2L		0x08
	#define LCDC_FUNC1L		0
	#define LCDC_FUNC5x10	0x4
	#define LCDC_FUNC5x7	0
#define LCDC_CGA		0x40
#define LCDC_DDA		0x80

//--------------------------------------------------
// Generowanie opónieñ
#define delay250ns() {asm volatile("nop"::);}

#define delayus8(t)\
	{asm volatile( \
		"delayus8_loop%=: \n\t"\
			"nop \n\t"\
			"dec %[ticks] \n\t"\
			"brne delayus8_loop%= \n\t"\
	: :[ticks]"r"(t) );}
	// DEC - 1 cykl, BRNE 2 cykle, + 1xnop. Zegar 4MHz

void delay100us8(uint8_t t)
{
	while(t>0)
	{
		delayus8(200); 
		--t; 
	}
}
void czekaj(unsigned long pt)
//funkcja opó¿nienia
//z ksi¹¿ki "Mikrokontrolery AVR w praktyce" Jaros³aw Doliñski wydawnictwo BTC Legionowo 2008
{
  #define tau 10.38
  unsigned char pt1;
  
  for(;pt>0;pt--)
  {
    for(pt1=0;pt1!=255;pt1++);
  }
}
// Koniec opónieñ
//--------------------------------------------------

//--------------------------------------------------
// Obs³uga wywietlacza
// Funkcje niskiego poziomu
#define LCD_EPULSE() \
	{PORT(LCD_EPORT) |= 1<<LCD_E; \
	delay250ns(); \
	PORT(LCD_EPORT) &= ~(1<<LCD_E);}

void LCDsendHalf(uint8_t data)
{
	data = (data & 0x0F); //zeruje 4 starsze bity 
	//data |= 0xF0;
	//data &= 0x0F;
	//data <<= LCD_D4; 
	PORT(LCD_DPORT) &= ~(0x0f << LCD_D4);  // wyzeruj 4 m³odsze bity
	PORT(LCD_DPORT) |= data<<LCD_D4;
	//PORT(LCD_DPORT) = PORT(LCD_DPORT) | (((data & 0x08)>>3) | ((data & 0x04)>>1) | ((data & 0x02)<<1) | ((data & 0x01)<<3));              
	//	(PORT(LCD_DPORT) & ~(0x0F<<LCD_D4)) | data;
	//PORT(LCD_DPORT) &= ~(0x0F<<LCD_D4); 
	//PORT(LCD_DPORT) |= data; 
	LCD_EPULSE();
}

void LCDsend(uint8_t data)
{
	// Starsza czêæ
	LCDsendHalf(data>>4);
	// M³odsza czêæ
	LCDsendHalf(data);
	delayus8(240);
}

// Funkcje interfejsu
void LCDcommand(uint8_t command)
{
	PORT(LCD_RSPORT) &= ~(1<<LCD_RS);
	LCDsend(command);
}

void LCDdata(uint8_t data)
{
	PORT(LCD_RSPORT) |= 1<<LCD_RS;
	LCDsend(data);
}

void LCDcls(void)
{
	LCDcommand(LCDC_CLS);
	delay100us8(96);
}

void LCDhome(void)
{
	LCDcommand(LCDC_HOME);
	delay100us8(96);
}

void LCDinit(void)
{
	delay100us8(150);
	delay100us8(150);
	PORT(LCD_RSPORT) &= ~(1<<LCD_RS);
	LCDsendHalf(LCDC_FUNC|LCDC_FUNC8b); 
	delay100us8(82);
	LCDsendHalf(LCDC_FUNC|LCDC_FUNC8b);
	delay100us8(4);
	LCDsendHalf(LCDC_FUNC|LCDC_FUNC4b);
	delay100us8(4);
	// Teraz jest ju¿ 4b. Koniec korzystania z sendHalf
	LCDcommand(LCDC_FUNC|LCDC_FUNC4b|LCDC_FUNC2L|LCDC_FUNC5x7);
	LCDcommand(LCDC_ON);
	LCDcls();
	LCDcommand(LCDC_MODE|LCDC_MODER);
	LCDcommand(LCDC_ON|LCDC_ONDISPLAY|LCDC_ONCURSOR);
}
void LCDstr(char* str)
{
 int n=0;
 
 while(0 != str[n])
 {
    LCDdata(str[n]);
	n++;
 }     
}
// Koniec obsługi wywietlacza
//--------------------------------------------------


//funkcja przelicza wartość z ADC na temperature w stopniach C
int getTemp(int a)
{
	float const Pt100[] = { 80.31,   82.29,  84.27,  86.25,  88.22,  90.19,  92.16,  94.12,  96.09,  98.04,
                            100.0,  101.95, 103.9,  105.85, 107.79, 109.73, 111.67, 113.61, 115.54, 117.47,
                            119.4,  121.32, 123.24, 125.16, 127.07, 128.98, 130.89, 132.8,  134.7,  136.6,
                            138.5,  140.39, 142.29, 157.31, 175.84, 195.84 };
    int t = -50, i = 0, dt = 0;
	float r = (100.0*a)/(1023-a);
    if (r > Pt100[0])
      while (250 > t) {
        dt = (t < 110) ? 5 : (t > 110) ? 50 : 40;
        if (r < Pt100[++i])
          return t + (r - Pt100[i-1]) * dt / (Pt100[i] - Pt100[i-1]);
        t += dt;
      };
 
    return (int)t;
}

//obsługa ADC

int getADC(char chanel)
//number należy do <0, 7>, kod nie sprawdza szkoda pamięci FLASH
{
  ADMUX &= 0xf0; //zerowanie 4 młodszych bitów
  ADMUX |= chanel; //ustawianie numeru kanału 
  
  ADCSRA |= 1<<ADSC; //zaczyna konwersję
  while(ADCSRA & (1 << ADSC));  //dopóki bit ADIF nie jest ustawiony nie jest gotowa kowersja
  ADCSRA &= ~((1 << ADSC) | (1 << ADIF)); //zeruj flage konwersji i rozkaz rozpoczecia
  
  ADCSRA |= 1<<ADSC; //zaczyna konwersję
  while(ADCSRA & (1 << ADSC));  //dopóki bit ADIF nie jest ustawiony nie jest gotowa kowersja
  ADCSRA &= ~((1 << ADSC) | (1 << ADIF)); //zeruj flage konwersji i rozkaz rozpoczecia
  return ADCW;
}
//pomiar predkosci obrotowej silnika
volatile int licz2, obroty;

ISR(TIMER2_OVF_vect)
{
//przerwanie od timera 2 -> odczyt liczby obrotów z T0
    TCNT2 = 256-250;
	licz2++;
	if(licz2 >= 125)
	{
		licz2 = 0;
		obroty = TCNT0;
		TCNT0 = 0;
	}
}

// START
int main(void)
{
	//konfiguracja portów
	DDRD = 1<<LCD_E | 1<<LCD_RS | (1<<2) | (1<<7);
	DDRB = (0x0F<<LCD_D4) | 1 << 0 | 1 << 1;

	//konfiguracja ADC
	ADMUX |= (0<<REFS0)|(0<<REFS1); //
	ADCSRA |= (1<<ADEN);

	//konfiguracja Timer1 jako PWM dla silnika
	TCCR1A |= 1 << COM1A1 | 1 << WGM11 | 1 << WGM10;
	TCCR1B |= 1 << WGM12 | 1 << CS10 | 1 << CS11;
	
	//konfiguracja Timera0 aby zliczał impulsy z enkodera
	TCCR0 |= (1<<CS02)|(1<<CS01); //zlicza z zewnątrz
	//przerwania od Timera2
	TCCR2 |= (1<<CS22)|(1<<CS21);
	TCNT2 = 256-250; 
	TIMSK |= 1<<TOIE2;
	//włącz globalnie przerwania
	sei();
	
	//ekran startowy
	LCDinit();
	LCDcls();
	LCDstr("  MIESZADELKO");
	LCDcommand(0xA8);
	LCDstr("          Kozik");
	_delay_ms(600);
		
	//zmienne
	int tg, t, predk, tz, u, us, du, tc;
	//temperatura grzałki, odczyt z potencjometru temperatury, odczyt z potencjometru prędkości, zadana temperatura, uchyb, poprzedni uchyb, zmiana uchybu, suma uchybów 
	u = 0;
	us = 0;
	tc = 0;
	char kierunek = 0;  //kierunek obrotu silnika
	int grzalka = 0;	//wypełnienie PWM grzałki
	int nri = 100;	//numer iteracji % 100 (do PWM grzałki)
	OCR1A = 0;	//PWM silnika
	
	//Pętla główna
    while(1)
	{
		//Wczytywanie danych z ADC
		tg = getTemp(getADC(0));
		t = getADC(2);
		predk = getADC(3);
		
		//ustawienia silnika
		if(predk < 460)
			if(kierunek != 1)
			{
				int temp = OCR1A;
				OCR1A = 0;
				PORTD |= 1<<2;	
				PORTB &= ~(1<<0);
				OCR1A = temp;
				kierunek = 1;					
			}
		if(predk > 540)
			if(kierunek != 2)
			{
				int temp = OCR1A;
				OCR1A = 0;
				PORTB |= 1<<0;
				PORTD &= ~(1<<2);
				kierunek = 2;
				OCR1A = temp;
			}
		if((predk >= 460) && (predk <= 540) && (kierunek != 0))
		{
			OCR1A = 0;
			PORTB &= ~(1<<0);
			PORTD &= ~(1<<2);
			kierunek = 0;
		}
		if(kierunek > 0)
		{
			if(predk < 30)	OCR1A = 1023;
			else if(predk < 512) OCR1A=1023-(predk*1.5);
			else if(predk > 960) OCR1A = 1023;
			else OCR1A = predk*0.8;
		}
		
		//odczyt temp zadanej
		if(t < 20)	tz = 0;
		else if(t > 1000) tz = 150;
		else tz = t*0.12+28;
		
		
		//PWM Grzałki	
		if(nri >= 100)
		{
			nri = 0;
			PORTD &= ~(1<<7);
			
			us = u;
			u = tz-tg;
			tc += u;
			du = u - us;
			if(tc < -20) tc = -20;
			
			//regulacja temperatury
			if(tz == 0)	
				grzalka = 0;
			else 
				grzalka = 0.4*u+0.3*du+0.1*tc;  //
		
			if(grzalka < 0)	grzalka = 0;
			if(grzalka > 100) grzalka = 100;
		}
		if(nri >= grzalka)
		{
			PORTD |= (1<<7);
		}
		
		
		//wyswietl dane na LCD
		if(nri%10 == 0)
		{
			char buffer1[16];
			char buffer2[16];
			sprintf(buffer1, "P=%3d%%   %4d Hz", grzalka, obroty/5);
			sprintf(buffer2, "%3d %cC <- %3d %cC", tg, 0xdf, tz, 0xdf);
			LCDcls();
			LCDstr(buffer2);
			LCDcommand(0xA8);
			LCDstr(buffer1);
		}
		
		nri++;
		_delay_ms(100);
		
		
	}
	return 0;
}
